div {
  content: compact(yellow, false, false, false, false);
  content: type-of(compact(yellow, false, false, false, false));
  content: length(type-of(compact(yellow, false, false, false, false)));
}