#one #two #three a, #one #two #three b {
  key: value;
}
#one #three c, #one #two #three d {
  key: value;
}
#one #two #three e, #one #two #three f {
  key: value;
}
#one.x #two.y #three.z g, #one.x #two.y #three.z h {
  key: value;
}
#one #two #three i, #one.x #two.y #three.z j {
  key: value;
}
#one #two k, #one #two #three l {
  key: value;
}
#two #three m, #one #two #three n {
  key: value;
}
#one #two #three o, #one #two #three #four p {
  key: value;
}
#one #two #three #four q, #one #two #three #four r {
  key: value;
}
#one #two #four s, #one #two #three #four t, #one #three #two #four t {
  key: value;
}
#one #three #five u, #one #two #three #five #four v,
#one #two #three #four #five v, #two #one #three #five #four v,
#two #one #three #four #five v {
  key: value;
}
#one #two #four #five w, #one #two #three #four #five x,
#one #three #two #four #five x {
  key: value;
}
one two one three one four one y, one two five one three one four one six z,
one two five one three one six four one z,
five one two one three one four one six z,
five one two one three one six four one z {
  key: value;
}
five one three one six aa, five one two one three one four one six bb,
one two five one three one four one six bb {
  key: value;
}
