@option style:legacy, short_colors:yes, reverse_colors:yes;
@variables {
    $path = unquote('../img/tmpl1/png');
    $color1 = #FF00FF;
}
.selector {
    background-image: url(#{$path}/head_bg.png);
    background-color: $color1;
    border-top: 1px solid $color1;
}
