$foo: foo;

.#{$foo} #{$foo}-#{$foo} {
    one#{$foo}three: $foo;

    // TODO Ruby Sass changes these single quotes into double quotes, which I
    // guess we want to preserve
    string: hello#{$foo}goodbye;
    string-single: 'he\'llo#{$foo}good"bye';
    string-double: "he\"llo#{$foo}good'bye";

    url: url(hello#{$foo}goodbye);
    url-single: url('he\'llo#{$foo}good"bye');
    url-double: url("he\"llo#{$foo}good'bye");

    nested: "#{"type-of(f#{"o"}o)" "bar"}";
}
