@option style:legacy;

p {
  color: #f00;
}
@-moz-keyframes spinnerRotate,
@-webkit-keyframes spinnerRotate,
@keyframes spinnerRotate {
 from {
   -webkit-transform: rotate(0deg);
   -moz-transform: rotate(0deg);
   transform: rotate(0deg);
  }
  75% {
    font-size: 300%;
    margin-left: 25%;
    width: 150%;
  }
 to {
   -webkit-transform: rotate(360deg);
   -moz-transform: rotate(360deg);
   transform: rotate(360deg);
  }
}
@-webkit-keyframes spinnerRotate {
  0% { top: 0; left: 0; }
  30% { top: 50px; }
  68%, 72% { left: 50px; }
  100% { top: 100px; left: 100%; }
}
a {
  color: #0f0;
}
