@option style:legacy, short_colors:yes, reverse_colors:yes;
.basicList {
    li {
        padding: 5px 10px;
        border-bottom: 1px solid #000000;
    }
    dd {
        margin: 4px;
    }
    span {
        display: inline-block;
    }
}
.roundBox {
    some: props;
}
.specialClass extends .basicList & .roundBox {}