.mod {
  background: url(static/assets/squares-n1XoWeNUWSLsX5d823Kf8w.png) no-repeat;
}
.mod.ten-by-ten {
  width: 10px;
  height: 10px;
  background-position: 0px 0px;
}
.mod.twenty-by-twenty {
  width: 20px;
  height: 20px;
  background-position: 0px -10px;
}
