.specialClass {
  padding: 10px;
  font-size: 14px;
}
.basicClass, .specialClass {
  padding: 20px;
  background-color: #FF0000;
}
.basicClass a, .specialClass a, .specialLink {
  text-decoration: none;
}
