div {
  background-image: image-url("test.png");
  fudge: walnut;
}
@media (-webkit-min-device-pixel-ratio: 2), (-moz-min-device-pixel-ratio: 2) {
  div {
    background-image: image-url("test@2x.png");
    -webkit-background-size: 100px 300px;
  }
}
span {
  blah: "foo bar";
  bleh: image-url("foo bar");
}
