div {
  $things: red 2/3 blue;
  content: $things;
  content: nth($things, 2);
  content: type-of(nth($things, 2));
  content: type-of(nth($things, 3));
  /**** #{2+2} ****/
  content: (1/2 3/4) + (5/6 7/8);
  /**** ****/
  @each $x in 1 2 3/4 {
    foo: $x;
    bar: $x + 1;
  }
}