@option style:legacy;

a {
  @each $name, $number in (foo: 1px, bar: 2px, baz: 3px) {
    #{$name}: $number;
  }
}

c {
  @each $key, $value in (foo bar) (bar, baz) bang {
    #{$key}: "Value is #{$value}";
  }
}
