p {
  color: #f00;
}
@media screen, print {
  body {
    line-height: 1.2;
  }
}
@media (min-width: 700px), handheld and (orientation: landscape) {
  body {
    background: #123;
  }
  div {
    background: #456;
  }
}
a {
  color: #0f0;
}
