@option style:legacy, warn_unused:yes;

@import "functions-context";

#test {
	@import "functions-context";
	@import "imports-unused";
	@import "imports-unused-stuff";
	width: grid-width(10);
}

div {
	width: grid-width(20);
}
