@option style:legacy;

$color: black;
.scoped {
    $bg: blue;
    $color: white;
    color: $color;
    background-color: $bg;
}
.unscoped {
    color: $color;
    // Would be an error
    // background: $bg;
}
