div {
  $list: append(1/2 3, 4);
  content: (1 2 3) == (1, 2, 3);
  content: (1 2 3) == (1 2 3);
  content: var $list;
  content: lit (1/2 3 4);
  content: (1/2 3 4) == $list;
  /***/
  content: length($list);
  content: type-of(nth($list, 1));
  content: nth($list, 1);
  content: nth(1/2 3 4, 1);
  $a: 1 2 3;
  $b: (1 2 3);
  content: $a == $b;
}