@option style:legacy, short_colors:yes, reverse_colors:yes;
@variables {
    $color = #FFF555;
}
.selector {
    padding: [5px * 2];
    color: [#ccc * 2];
    // lets assume $color is '#FFF555'
    background-color: [$color - #222 + #101010];
}
