@option style:legacy, short_colors:yes, reverse_colors:yes;
// This ruleset won't be rendered on its own.
#context a%extreme {
  color: blue;
  font-weight: bold;
  font-size: 2em;
}
%unused {
  foo: bar;
}
.notice { @extend %extreme; }
