@option style:legacy, short_colors:yes, reverse_colors:yes;
/* style.scss */
.fakeshadow {
  border: {
    style: solid;
    left: {
      width: 4px;
      color: #888;
    }
    right: {
      width: 2px;
      color: #ccc;
    }
  }
}