@option style:legacy, short_colors:yes, reverse_colors:yes;
@variables {
    $ie6 = unquote("*html");
    $ie7 = unquote("*:first-child+html");
}
$ie6 {
    .a  { color:#fff; }
    .b  { color:#000; }
}
$ie7 {
    .a  { color:#fff; }
    .b  { color:#000; }
}
