@option style:legacy;

@import "lib";
@import "part";

div {
    @include transition(opacity $transition-time linear, margin $transition-time linear);
}

h1 {
    @include junk;  // This mixin does not exist intentionally!
}
