div {
  span, p, span {
    color: red;
  }
  a.foo.bar.foo {
    color: green;
  }
  &:nth(-3) {
    color: blue;
  }
}

@-webkit-keyframes {
  from {
    left: 0px;
  }
  to {
    left: 200px;
  }
}