@option style:legacy;

$images: sprite-map("squares/*.png", $cache-buster: false);

.mod {
  background: $images;
  &.ten-by-ten {
    $file: sprite-file($images, "ten-by-ten");
    width: image-width($file);
    height: image-height($file);
    background-position: sprite-position($images, "ten-by-ten");
  }
  &.twenty-by-twenty {
    $file: sprite-file($images, "twenty-by-twenty");
    width: image-width($file);
    height: image-height($file);
    background-position: sprite-position($images, "twenty-by-twenty");
  }
}
